	PAGE	60,132
        TITLE   Tandy/PCjr 1000 Sound Extensions
	NAME	TANDYSND

;------------------------------------------------------------------------------;
;									       ;
;                                TANDYSND 1.2                                  ;
;									       ;
;                       Tandy/PCjr 1000 Sound Extensions                       ;
;									       ;
;									       ;
;				  Written By:				       ;
;			      Joseph A. Albrecht			       ;
;									       ;
;------------------------------------------------------------------------------;
PUSHR	MACRO

	PUSH	AX
	PUSH	BX
	PUSH	CX
	PUSH	DX
	PUSH	SI
	PUSH	DI
	PUSH	BP

	ENDM

POPR	MACRO

	POP	BP
	POP	DI
	POP	SI
	POP	DX
	POP	CX
	POP	BX
	POP	AX

	ENDM

;Stack frame equates
SS_AX		EQU	WORD PTR [BP+16]
SS_BX		EQU	WORD PTR [BP+14]
SS_CX		EQU	WORD PTR [BP+12]
SS_DX		EQU	WORD PTR [BP+10]
SS_SI		EQU	WORD PTR [BP+08]
SS_DI		EQU	WORD PTR [BP+06]
SS_BP		EQU	WORD PTR [BP+04]
SS_DS		EQU	WORD PTR [BP+02]
SS_ES		EQU	WORD PTR [BP+00]

;High/Low values to validate function calls
FunctLo		EQU	80H
FunctHi		EQU	86H

;Equates for 'Noise' and 'Sound' procedures
BufSize		EQU	64
BufEntry	EQU	6

Code	SEGMENT PARA PUBLIC 'Code'

	ASSUME	CS:Code,DS:Code,ES:Code,SS:Code

	ORG	100H

Start:
	JMP	Setup		 ;Jump to TSR initialization code

	EVEN

OldINT10	LABEL	DWORD	;Symbolic name to execute old interrupt 10H
INT10		DW	2 DUP(?);Old video BIOS interrupt vector address
OldINT08	LABEL	DWORD	;Symbolic name to execute old interrupt 08H
INT08		DW	2 DUP(?);Old BIOS timer interrupt vector address
OldINT08X	LABEL	DWORD	;Symbolic name to execute old interrupt 08H
INT08X		DW	2 DUP(?);Old BIOS timer interrupt vector address
TimerSpeed	DW	0	;Speed of 8253 timer chip flag
TimerValue	DW	0	;8253 timer chip speed value
TimerCount	DW	0	;Number of interrupts before executing INT 08H
TimerFlag	DW	0	;Timer interrupt status flag:
				;  Bit 0  - Sound channel 0 active
				;  Bit 1  - Sound channel 1 active
				;  Bit 2  - Sound channel 2 active
				;  Bit 3  - Noise channel active
				;  Bit 4  - Sound channel 0 busy
				;  Bit 5  - Sound channel 1 busy
				;  Bit 6  - Sound channel 2 busy
				;  Bit 7  - Noise channel busy
				;  Bit 8  - Sound buffering: On/Off
				;  Bit 9  - Noise buffering: On/Off
				;  Bit 10 - Delay Active
				;  Bit 11 - Skip Timer INT 08H routine
SoundCnt0	DW	?	;Clock ticks before turning off channel 0
SoundCnt1	DW	?	;Clock ticks before turning off channel 1
SoundCnt2	DW	?	;Clock ticks before turning off channel 2
NoiseCnt	DW	?	;Clock ticks before turning off noise channel
DelayCnt	DW	?	;Clock ticks before quitting delay loop
PBufPtr1	DW OFFSET SoundBuf1 ;Sound buffer 1 pointer for putting sounds
PBufPtr2	DW OFFSET SoundBuf2 ;Sound buffer 2 pointer for putting sounds
PBufPtr3	DW OFFSET SoundBuf3 ;Sound buffer 3 pointer for putting sounds
PBufPtr4	DW OFFSET NoiseBuf  ;Noise buffer pointer for putting noises
GBufPtr1	DW OFFSET SoundBuf1 ;Sound buffer 1 pointer for getting sounds
GBufPtr2	DW OFFSET SoundBuf2 ;Sound buffer 2 pointer for getting sounds
GBufPtr3	DW OFFSET SoundBuf3 ;Sound buffer 3 pointer for getting sounds
GBufPtr4	DW OFFSET NoiseBuf  ;Noise buffer pointer for getting noises
BufCnt1		DW	0	;Sound buffer 1 current total
BufCnt2		DW	0	;Sound buffer 2 current total
BufCnt3		DW	0	;Sound buffer 3 current total
BufCnt4		DW	0	;Noise buffer current total

;Buffer 1 - Sound channel 0
SoundBuf1	DW  BufSize * BufEntry / 2 DUP (?)
BufEnd1:			;End of buffer 1
;Buffer 2 - Sound channel 1
SoundBuf2	DW  BufSize * BufEntry / 2 DUP (?)
BufEnd2:			;End of buffer 2
;Buffer 3 - Sound channel 2
SoundBuf3	DW  BufSize * BufEntry / 2 DUP (?)
BufEnd3:			;End of buffer 3
;Buffer 4 - Noise channel
NoiseBuf	DW  BufSize * BufEntry / 2 DUP (?)
BufEnd4:			;End of buffer 4

SoundTab	DW	Sound0	;Sound channel table
		DW	Sound1	; ..
		DW	Sound2	; ..

;CALL table for functions
Function	DW	System		;80H - Global system functions
		DW	Sound		;81H - Generate a sound
		DW	Noise		;82H - Generate a noise
		DW	Delay		;83H - Delay loop
		DW	FastTimer	;84H - Set timer chip to fast speed
		DW	ResetTimer	;85H - Set timer chip to normal speed
		DW	GetTimer	;86H - Get timer chip speed

TandySnd:
	CMP	AH,0EEH		;Check for a function call
	JNE	ExecINT10	; ..
	CMP	AL,FunctLo	;Validate function number
	JB	ExecINT10	; ..
	CMP	AL,FunctHi	; ..
	JA	ExecINT10	; ..
	STI			;Keep interrupts going
	PUSHR			;Save registers
	PUSH	DS		; ..
	PUSH	ES		; ..
	MOV	BP,CS		;Establish Data Segment for INT 10H extension
	MOV	DS,BP		; ..
	SUB	AL,FunctLo	   ;Index into function table
	CBW			   ; ..
	SHL	AX,1		   ; ..
	MOV	BP,AX		   ; ..
	MOV	AX,CS:Function[BP] ; ..
	MOV	BP,SP		;Set up stack frame reference
	CALL	AX		;CALL requested function
	POP	ES		;Restore registers
	POP	DS		; ..
	POPR			; ..
	IRET			;Return to interrupted program

ExecINT10:
	JMP	CS:[OldINT10]	;Execute the normal video BIOS

;
;Return residency status - Set audio system
;

System:
	OR	BX,BX		;Status check?
	JNZ	System1		; ..
	MOV	SS_AX,0DEADH	;Store in AX position on stack
	RET
System1:
	CMP	BX,4		;Turn all current sound processing off?
	JNE	System2		; ..
	CALL	SoundOff	;Reset sound system
	RET
System2:
	CMP	BX,5		;Turn sound buffering on?
	JNE	System4		; ..
	TEST	[TimerFlag],100H;See if buffering is already on
	JNZ	System3		; ..
	OR	[TimerFlag],100H;Turn all sound buffer processing on
	CALL	ResetSoundBuf	;Reset sound buffer variables
System3:
	RET
System4:
	CMP	BX,6		;Turn sound buffering off?
	JNE	System5		; ..
	AND	[TimerFlag],0FE8FH ;Turn all sound buffer processing off
	RET
System5:
	CMP	BX,7		;Turn all noise processing off?
	JNE	System6		; ..
	CALL	NoiseOff	;Reset noise system
	RET
System6:
	CMP	BX,8		;Turn noise buffering on?
	JNE	System8		; ..
	TEST	[TimerFlag],200H;See if buffering is already on
	JNZ	System7		; ..
	OR	[TimerFlag],200H;Turn all noise buffer processing on
	CALL	ResetNoiseBuf	;Reset noise buffer variables
System7:
	RET
System8:
	CMP	BX,9		;Turn noise buffering off?
	JNE	System9		; ..
	AND	[TimerFlag],0FD7FH ;Turn all noise buffer processing off
	RET
System9:
	CMP	BX,10		;Turn all sound & noise processing off?
	JNE	SystemExit	; ..
	CALL	SoundOff	;Reset sound system
	CALL	NoiseOff	;Reset noise system
SystemExit:
	RET

;
;Turns all current sound processing off
;

SoundOff:
	AND	[TimerFlag],0FE88H ;Turn all sound buffer processing off
	CALL	Sound0Off	;Turn off all sound channels
	CALL	Sound1Off	; ..
	CALL	Sound2Off	; ..
	CALL	ResetSoundBuf	;Reset sound buffers
	RET

;
;Turn sound channel 0 off
;

Sound0Off:
	PUSH	AX		;Save AX
	MOV	AX,0083H	;Turn off sound channel 0
        OUT     0C0H,AL         ; ..
        MOV     AL,AH           ; ..
        OUT     0C0H,AL         ; ..
	POP	AX		;Restore AX
	RET

;
;Turn sound channel 1 off
;

Sound1Off:
	PUSH	AX		;Save AX
	MOV	AX,00A3H	;Turn off sound channel 1
        OUT     0C0H,AL         ; ..
        MOV     AL,AH           ; ..
        OUT     0C0H,AL         ; ..
	POP	AX		;Restore AX
	RET

;
;Turn sound channel 2 off
;

Sound2Off:
	PUSH	AX		;Save AX
	MOV	AX,00C3H	;Turn off sound channel 2
        OUT     0C0H,AL         ; ..
        MOV     AL,AH           ; ..
        OUT     0C0H,AL         ; ..
	POP	AX		;Restore AX
	RET

;
;Turns all current noise processing off
;

NoiseOff:
	AND	[TimerFlag],0FD77H ;Turn all noise buffer processing off
	CALL	NoiseOffX	;Turn noise channel off
	CALL	ResetNoiseBuf	;Reset noise buffer
	RET

;
;Turn noise channel off
;

NoiseOffX:
	PUSH	AX		;Save AX
	MOV	AL,0FFH		;Turn noise channel off
	OUT	0C0H,AL		; ..
	POP	AX		;Restore AX
	RET

;
;Reset all sound buffer variables
;

ResetSoundBuf:
	MOV	[BufCnt1],0    ;Reset all counters
	MOV	[BufCnt2],0    ; ..
	MOV	[BufCnt3],0    ; ..
	MOV	[GBufPtr1],OFFSET SoundBuf1 ;Reset all pointers
	MOV	[PBufPtr1],OFFSET SoundBuf1 ; ..
	MOV	[GBufPtr2],OFFSET SoundBuf2 ; ..
	MOV	[PBufPtr2],OFFSET SoundBuf2 ; ..
	MOV	[GBufPtr3],OFFSET SoundBuf3 ; ..
	MOV	[PBufPtr3],OFFSET SoundBuf3 ; ..
	RET

;
;Reset all noise buffer variables
;

ResetNoiseBuf:
	MOV	[BufCnt4],0    ;Reset counter
	MOV	[GBufPtr4],OFFSET NoiseBuf ;Reset all pointers
	MOV	[PBufPtr4],OFFSET NoiseBuf ; ..
	RET

;
;Produce a sound with the specified frequency, duration, volume, sound channel
;

Sound:
	OR	DI,DI		;Check for zero length sound
	JNZ	SoundA		; ..
	RET
SoundA:
	AND	BX,15		;Make sure volume is 0-15
	AND	CX,3		;Make sure sound channel is 0-2
	CMP	SI,110		;Make sure frequency is at least 110Hz
	JAE	SoundB		; ..
	MOV	SI,110		; ..
SoundB:
	MOV	BP,CX		;Jump to selected sound channel routine
	SHL	BP,1		; ..
	JMP	CS:SoundTab[BP] ; ..

;Sound channel 0
Sound0:
	TEST	[TimerFlag],1	;Check if current sound still active
	JZ	Sound0B		; ..
	OR	[TimerFlag],10H ;Turn on busy flag bit
	TEST	[TimerFlag],100H;See if sound buffer is active
	JZ	Sound0A		; ..
	JMP	PutSound1	; ..
Sound0A:
	CMP	[SoundCnt0],0	;Keep looping until time has expired before
	JA	Sound0A		; .. making next sound
Sound0B:
	MOV	[SoundCnt0],DI	;Set number of ticks for channel 0
	OR	[TimerFlag],1	;Turn channel 0 flag bit on
	JMP	MakeSound	;Play the sound

;Sound channel 1
Sound1:
	TEST	[TimerFlag],2	;Check if current sound still active
	JZ	Sound1B		; ..
	OR	[TimerFlag],20H ;Turn on busy flag bit
	TEST	[TimerFlag],100H;See if sound buffer is active
	JZ	Sound1A		; ..
	JMP	PutSound2	; ..
Sound1A:
	CMP	[SoundCnt1],0	;Keep looping until time has expired before
	JA	Sound1A		; .. making next sound
Sound1B:
	MOV	[SoundCnt1],DI	;Set number of ticks for channel 0
	OR	[TimerFlag],2	;Turn channel 1 flag bit on
	JMP	MakeSound	;Play the sound

;Sound channel 2
Sound2:
	TEST	[TimerFlag],4	;Check if current sound still active
	JZ	Sound2B		; ..
	OR	[TimerFlag],40H ;Turn on busy flag bit
	TEST	[TimerFlag],100H;See if sound buffering is active
	JZ	Sound2A		; ..
	JMP	PutSound3	; ..
Sound2A:
	CMP	[SoundCnt2],0	;Keep looping until time has expired before
	JA	Sound2A		; .. making next sound
Sound2B:
	MOV	[SoundCnt2],DI	;Set number of ticks for channel 0
	OR	[TimerFlag],4	;Turn channel 2 flag bit on

;Get the current sound channel volume
MakeSound:
	MOV	BH,CL		;Sound volume formula:
	MOV	CL,5		; .. 90H OR (Channel * 32) OR (Volume XOR 15)
	SHL	BH,CL		; ..
	MOV	CH,BL		; ..
	XOR	CH,15		; ..
	OR	CH,BH		; ..
	OR	CH,90H		; ..

;Get the current sound channel tone (2 bytes)
	MOV	DX,1		;Calculate the tone (111860 / Frequency)
	MOV	AX,0B4F4H	; ..
	DIV	SI		; ..
	MOV	DX,AX		; ..
	AND	AL,15		;Byte 1 formula:
	OR	AL,BH		; .. 80H OR (Channel * 32) OR (Tone AND 15)
	OR	AL,80H		; ..
	AND	DX,3F0H		;Byte 2 formula:
	MOV	CL,4		; .. (Tone AND 3F0H) / 16
	SHR	DX,CL		; ..

;Play the sound
        OUT     0C0H,AL         ;Set the tone
        MOV     AL,DL           ; ..
        OUT     0C0H,AL         ; ..
	MOV	AL,CH		;Set the volume
	OUT	0C0H,AL		; ..
	RET

;
;Generate a periodic or white noise with the specified volume and duration
;

Noise:
	OR	DX,DX		;Check for zero length noise
	JZ	Noise3		; ..
	TEST	[TimerFlag],8	;See if current noise still active
	JZ	Noise2		; ..
	OR	[TimerFlag],80H ;Turn on busy flag
	TEST	[TimerFlag],200H;See if buffering active
	JZ	Noise1		; ..
	JMP	PutNoise	;Put noise into sound buffer
Noise1:
	CMP	[NoiseCnt],0	;Keep looping until time has expired
	JA	Noise1		; ..
Noise2:
	MOV	[NoiseCnt],DX	;Set duration
	OR	[TimerFlag],8	;Turn noise channel flag bit on

;Get noise volume
	MOV	AH,CL		;Noise volume formula:
	AND	AH,15		; .. 0F0H OR ((Volume AND 15) XOR 15)
	XOR	AH,15		; ..
	OR	AH,0F0H		; ..

;Get noise type
	MOV	AL,BL		;Noise type formula:
	MOV	CL,2		; .. 0E0H OR ((Type / 4) * 4) OR (Type AND 3)
	AND	AL,3		; ..
	SHR	BL,CL		; ..
	SHL	BL,CL		; ..
	OR	AL,BL		; ..
	OR	AL,0E0H		; ..

;Make the noise
        OUT     0C0H,AL         ;Set the type
        MOV     AL,AH           ;Set the volume
        OUT     0C0H,AL         ; ..
Noise3:
	RET

;
;Put the current sound into the sound buffer
;

;Sound buffer number 1
PutSound1:
	CMP	[BufCnt1],BufSize ;Keep looping until there is an opening in
	JE	PutSound1	  ; .. the buffer
	PUSH	BP		;Save BP
	MOV	BP,[PBufPtr1]	    ;See if at end of buffer
	CMP	BP,OFFSET BufEnd1   ; ..
	JNE	PutSound1A	    ; ..
	MOV	BP,OFFSET SoundBuf1 ;Point to beginning of buffer
PutSound1A:
	MOV	CS:[BP],BX	;Store the volume
	MOV	CS:[BP+2],SI	;Store the frequency
	MOV	CS:[BP+4],DI	;Store the duration
	ADD	BP,BufEntry	;Increment buffer pointer
	MOV	[PBufPtr1],BP	; ..
	INC	[BufCnt1]	;Increment current sound total
	POP	BP		;Restore BP
	RET

;Sound buffer number 2
PutSound2:
	CMP	[BufCnt2],BufSize ;Keep looping until there is an opening in
	JE	PutSound2	  ; .. the buffer
	PUSH	BP		;Save BP
	MOV	BP,[PBufPtr2]	    ;See if at end of buffer
	CMP	BP,OFFSET BufEnd2   ; ..
	JNE	PutSound2A	    ; ..
	MOV	BP,OFFSET SoundBuf2 ;Point to beginning of buffer
PutSound2A:
	MOV	CS:[BP],BX	;Store the volume
	MOV	CS:[BP+2],SI	;Store the frequency
	MOV	CS:[BP+4],DI	;Store the duration
	ADD	BP,BufEntry	;Increment buffer pointer
	MOV	[PBufPtr2],BP	; ..
	INC	[BufCnt2]	;Increment current sound total
	POP	BP		;Restore BP
	RET

;Sound buffer number 3
PutSound3:
	CMP	[BufCnt3],BufSize ;Keep looping until there is an opening in
	JE	PutSound3	  ; .. the buffer
	PUSH	BP		;Save BP
	MOV	BP,[PBufPtr3]	    ;See if at end of buffer
	CMP	BP,OFFSET BufEnd3   ; ..
	JNE	PutSound3A	    ; ..
	MOV	BP,OFFSET SoundBuf3 ;Point to beginning of buffer
PutSound3A:
	MOV	CS:[BP],BX	;Store the volume
	MOV	CS:[BP+2],SI	;Store the frequency
	MOV	CS:[BP+4],DI	;Store the duration
	ADD	BP,BufEntry	;Increment buffer pointer
	MOV	[PBufPtr3],BP	; ..
	INC	[BufCnt3]	;Increment current sound total
	POP	BP		;Restore BP
	RET

;
;Put the current noise into the noise buffer
;

PutNoise:
	CMP	[BufCnt4],BufSize ;Keep looping until there is an opening in
	JE	PutNoise	  ; .. the buffer
	PUSH	BP		;Save BP
	MOV	BP,[PBufPtr4]	   ;See if at end of buffer
	CMP	BP,OFFSET BufEnd4  ; ..
	JNE	PutNoise1	   ; ..
	MOV	BP,OFFSET NoiseBuf ;Point to beginning of buffer
PutNoise1:
	MOV	CS:[BP],BX	;Store the type
	MOV	CS:[BP+2],CX	;Store the volume
	MOV	CS:[BP+4],DX	;Store the duration
	ADD	BP,BufEntry	;Increment buffer pointer
	MOV	[PBufPtr4],BP	; ..
	INC	[BufCnt4]	;Increment current sound total
	POP	BP		;Restore BP
	RET

;
;Get the current sound from the sound buffer and generate it
;

;Sound buffer number 1
GetSound1:
	CMP	[BufCnt1],0	;See if buffer is empty
	JE	GetSound1B	; ..
	PUSHR			;Save registers
	MOV	BP,[GBufPtr1]	    ;See if at end of buffer
	CMP	BP,OFFSET BufEnd1   ; ..
	JNE	GetSound1A	    ; ..
	MOV	BP,OFFSET SoundBuf1 ;Point to beginning of buffer
GetSound1A:
	MOV	BX,CS:[BP]	;Get the volume
	MOV	SI,CS:[BP+2]	;Get the frequency
	MOV	DI,CS:[BP+4]	;Get the duration
	MOV	CL,0		;Get the voice
	ADD	BP,BufEntry	;Increment buffer pointer
	MOV	[GBufPtr1],BP	; ..
	DEC	[BufCnt1]	;Decrement current sound total
	CALL	Sound		;Play the sound
	POPR			;Restore registers
GetSound1B:
	RET

;Sound buffer number 2
GetSound2:
	CMP	[BufCnt2],0	;See if buffer is empty
	JE	GetSound2B	; ..
	PUSHR			;Save registers
	MOV	BP,[GBufPtr2]	    ;See if at end of buffer
	CMP	BP,OFFSET BufEnd2   ; ..
	JNE	GetSound2A	    ; ..
	MOV	BP,OFFSET SoundBuf2 ;Point to beginning of buffer
GetSound2A:
	MOV	BX,CS:[BP]	;Get the volume
	MOV	SI,CS:[BP+2]	;Get the frequency
	MOV	DI,CS:[BP+4]	;Get the duration
	MOV	CL,1		;Get the voice
	ADD	BP,BufEntry	;Increment buffer pointer
	MOV	[GBufPtr2],BP	; ..
	DEC	[BufCnt2]	;Decrement current sound total
	CALL	Sound		;Play the sound
	POPR			;Restore registers
GetSound2B:
	RET

;Sound buffer number 3
GetSound3:
	CMP	[BufCnt3],0	;See if buffer is empty
	JE	GetSound3B	; ..
	PUSHR			;Save registers
	MOV	BP,[GBufPtr3]	    ;See if at end of buffer
	CMP	BP,OFFSET BufEnd3   ; ..
	JNE	GetSound3A	    ; ..
	MOV	BP,OFFSET SoundBuf3 ;Point to beginning of buffer
GetSound3A:
	MOV	BX,CS:[BP]	;Get the volume
	MOV	SI,CS:[BP+2]	;Get the frequency
	MOV	DI,CS:[BP+4]	;Get the duration
	MOV	CL,2		;Get the voice
	ADD	BP,BufEntry	;Increment buffer pointer
	MOV	[GBufPtr3],BP	; ..
	DEC	[BufCnt3]	;Decrement current sound total
	CALL	Sound		;Play the sound
	POPR			;Restore registers
GetSound3B:
	RET

;
;Get the current noise from the noise buffer and generate it
;

GetNoise:
	CMP	[BufCnt4],0	;See if buffer is empty
	JE	GetNoise2	; ..
	PUSHR			;Save registers
	MOV	BP,[GBufPtr4]	   ;See if at end of buffer
	CMP	BP,OFFSET BufEnd4  ; ..
	JNE	GetNoise1	   ; ..
	MOV	BP,OFFSET NoiseBuf ;Point to beginning of buffer
GetNoise1:
	MOV	BX,CS:[BP]	;Get the type
	MOV	CX,CS:[BP+2]	;Get the volume
	MOV	DX,CS:[BP+4]	;Get the duration
	ADD	BP,BufEntry	;Increment buffer pointer
	MOV	[GBufPtr4],BP	; ..
	DEC	[BufCnt4]	;Decrement current sound total
	CALL	Noise		;Make the noise
	POPR			;Restore registers
GetNoise2:
	RET

;
;Delay for the specified number of clock ticks
;

Delay:
	OR	BX,BX		;Make sure delay is greater than zero
	JZ	Delay2		; ..
	OR	[TimerFlag],400H;Turn delay flag bit on
	MOV	[DelayCnt],BX	;Set number of ticks for delay
Delay1:
	CMP	[DelayCnt],0	;Keep looping until time has expired
	JA	Delay1		; ..
Delay2:
	RET

;
;Set the 8253 timer chip for 2 - 256 times it's normal speed
;

FastTimer:
	CMP	BX,2		;Validate speed
	JB	FastTimer1	; ..
	CMP	BX,256		; ..
	JA	FastTimer1	; ..
	CALL	ResetTimer	;Make sure timer is reset to restore INT vector
	MOV	DX,1		;Get the timer chip speed value
	XOR	AX,AX		; .. (65536 / Speed)
	DIV	BX		; ..
	MOV	[TimerSpeed],BX ;Set the 8253 timer variable values
	MOV	[TimerCount],BX ; ..
	MOV	[TimerValue],AX ; ..
	CLI			;Disable interrupts
	XOR	AX,AX		;Set new INT 08H vector address
	MOV	ES,AX		; ..
	MOV	AX,ES:[20H]	; ..
	MOV	[INT08X],AX	; ..
	MOV	AX,ES:[22H]	; ..
	MOV	[INT08X+2],AX	; ..
	MOV	AX,OFFSET Timer ; ..
	MOV	ES:[20H],AX	; ..
	MOV	ES:[22H],CS	; ..
	CALL	SetTimer	;The the 8253 timer chip speed
	STI			;Re-enable interrupts
FastTimer1:
	RET

;
;Reset the 8253 timer chip to normal speed
;

ResetTimer:
	MOV	[TimerValue],0	;Set the 8253 timer chip to normal speed
ResetTimer1:
	CMP	[TimerSpeed],0	;Keep looping until speed is set to normal
	JNE	ResetTimer1	; .. by interrupt handler
	RET

;
;Return the 8253 timer chip speed
;

GetTimer:
	MOV	AX,[TimerSpeed] ;Return timer chip speed in BX
	OR	AX,AX		;See if it's 0, if so return 1
	JNZ	GetTimer1	; ..
	INC	AX		; ..
GetTimer1:
	MOV	SS_BX,AX	; ..
	RET

;
;This internal routine takes over the BIOS timer interrupt and updates the
;number of clock ticks that have elapsed since the start of the last sound,
;noise, or delay function. When the specified number of clock ticks has passed
;the sound or noise is turned off and the sound buffer is checked to see if if
;another sound or noise is waiting to be played. In the case of the delay
;function, control is returned to the interrupted program.
;

Timer:
	PUSH	AX		;Save registers
	PUSH	DS		; ..
	MOV	AX,CS		;Set up local data segment
	MOV	DS,AX		; ..
	TEST	[TimerFlag],800H ;See if we should skip interrupt
	JZ	Timer1		 ; .
	CALL	ResetSkip	;Reset interrupt skip variables
	JMP	Timer3		;Exit interrupt
Timer1:
	MOV	AX,[TimerFlag]	;Store timer flag for testing
	TEST	AX,40FH		;See if there is any processing to do
	JZ	Timer2		; ..
	CALL	ChkSnd0		;Check sound channel 0
	CALL	ChkSnd1		;Check sound channel 1
	CALL	ChkSnd2		;Check sound channel 0
	CALL	ChkNoise	;Check noise channel
	CALL	ChkDelay	;Check delay
Timer2:
	CMP	[TimerSpeed],0	;See if timer speed is normal
	JNE	Timer4		; ..
Timer3:
	POP	DS		;Restore registers
	POP	AX		; ..
	JMP	CS:[OldINT08]	;Execute normal BIOS timer interrupt
Timer4:
	DEC	[TimerCount]	;See if it's time to execute INT 08H
	JNZ	Timer5		; ..
	OR	[TimerFlag],800H;Skip this routine next time around
	POP	DS		;Restore registers
	POP	AX		; ..
	JMP	CS:[OldINT08X]	;Execute saved BIOS timer interrupt
Timer5:
	MOV	AL,20H		;Tell interrupt controller we're done here
	OUT	20H,AL		; ..
	POP	DS		;Restore registers
	POP	AX		; ..
	IRET			;Return from interrupt

;
;Reset skip interrupt variables
;

ResetSkip:
	AND	[TimerFlag],NOT 800H ;Reset skip flag
	MOV	AX,[TimerSpeed] ;Reset timer counter
	MOV	[TimerCount],AX ; ..
	CMP	[TimerValue],0	;See if timer speed should be set to normal
	JNE	ResetSkip1	; ..
	CLI			;Disable interrupts
	PUSH	ES		;Restore old INT 08H vector address
	XOR	AX,AX		; ..
	MOV	ES,AX		; ..
	MOV	AX,[INT08X]	; ..
	MOV	ES:[20H],AX	; ..
	MOV	AX,[INT08X+2]	; ..
	MOV	ES:[22H],AX	; ..
	POP	ES		; ..
	MOV	[TimerSpeed],0	;Reset timer speed to normal
	CALL	SetTimer	; ..
	STI			;Re-enable interrupts
ResetSkip1:
	RET

;
;Check sound channel 0
;

ChkSnd0:
	TEST	AL,1		;Check sound channel 0
	JZ	ChkSnd0C	; ..
	DEC	[SoundCnt0]	;See if time has expired
	JNZ	ChkSnd0C	; ..
	AND	[TimerFlag],NOT 1 ;Turn channel 0 flag bit off
	TEST	AL,10H		;See if another sound is waiting to play
	JNZ	ChkSnd0A	; ..
	CALL	Sound0Off	;Turn sound channel 0 off
ChkSnd0A:
	TEST	AX,100H		;See if sound buffer is active
	JZ	ChkSnd0B	; ..
	CALL	GetSound1	;Get next sound to play from the buffer
	CMP	[BufCnt1],0	;See if busy flag should be reset
	JNE	ChkSnd0C	; ..
ChkSnd0B:
	AND	[TimerFlag],NOT 10H ;Reset sound channel 0 busy flag
ChkSnd0C:
	RET

;
;Check sound channel 1
;

ChkSnd1:
	TEST	AL,2		;Check sound channel 1
	JZ	ChkSnd1C	; ..
	DEC	[SoundCnt1]	;See if time has expired
	JNZ	ChkSnd1C	; ..
	AND	[TimerFlag],NOT 2 ;Turn channel 1 flag bit off
	TEST	AL,20H		;See if another sound is waiting to play
	JNZ	ChkSnd1A	; ..
	CALL	Sound1Off	;Turn sound channel 1 off
ChkSnd1A:
	TEST	AX,100H		;See if sound buffer is active
	JZ	ChkSnd1B	; ..
	CALL	GetSound2	;Get next sound to play from the buffer
	CMP	[BufCnt2],0	;See if busy flag should be reset
	JNE	ChkSnd1C	; ..
ChkSnd1B:
	AND	[TimerFlag],NOT 20H ;Reset sound channel 1 busy flag
ChkSnd1C:
	RET

;
;Check sound channel 2
;

ChkSnd2:
	TEST	AL,4		;Check sound channel 2
	JZ	ChkSnd2C	; ..
	DEC	[SoundCnt2]	;See if time has expired
	JNZ	ChkSnd2C	; ..
	AND	[TimerFlag],NOT 4 ;Turn channel 2 flag bit off
	TEST	AL,40H		;See if another sound is waiting to play
	JNZ	ChkSnd2A	; ..
	CALL	Sound2Off	;Turn sound channel 2 off
ChkSnd2A:
	TEST	AX,100H		;See if sound buffer is active
	JZ	ChkSnd2B	; ..
	CALL	GetSound3	;Get next sound to play from the buffer
	CMP	[BufCnt3],0	;See if busy flag should be reset
	JNE	ChkSnd2C	; ..
ChkSnd2B:
	AND	[TimerFlag],NOT 40H ;Reset sound channel 2 busy flag
ChkSnd2C:
	RET

;
;Check noise channel
;

ChkNoise:
	TEST	AL,8		;Check noise channel
	JZ	ChkNoise3	; ..
	DEC	[NoiseCnt]	;See if time has expired
	JNZ	ChkNoise3	; ..
	AND	[TimerFlag],NOT 8 ;Turn noise flag bit off
	TEST	AL,80H		;See if another noise is waiting
	JNZ	ChkNoise1	; ..
	CALL	NoiseOffX	;Turn noise off
ChkNoise1:
	TEST	AX,200H		;See if noise buffer is active
	JZ	ChkNoise2	; ..
	CALL	GetNoise	;Get next noise to make from the buffer
	CMP	[BufCnt4],0	;See if busy flag should be reset
	JNE	ChkNoise3	; ..
ChkNoise2:
	AND	[TimerFlag],NOT 80H ;Reset noise channel busy flag
ChkNoise3:
	RET

;
;Check delay
;

ChkDelay:
	TEST	AX,400H		;Check delay
	JZ	ChkDelay1	; ..
	DEC	[DelayCnt]	;Decrement delay counter
	JNZ	ChkDelay1	; ..
	AND	[TimerFlag],NOT 400H ;Turn delay flag bit off
ChkDelay1:
	RET

;
;Set the speed of the 8253 timer chip
;

SetTimer:
	MOV	AL,36H		;Program 8253 command register
	OUT	43H,AL		; ..
	MOV	AX,[TimerValue] ;Set 8253 timer speed (Channel 0)
	OUT	40H,AL		; ..
	MOV	AL,AH		; ..
	OUT	40H,AL		; ..
	RET

;Check if computer is a Tandy 1000 series
Setup:
        MOV     AX,0FFFFH               ;Check byte at FFFF:000E for 0FDH/0FFH
        MOV     ES,AX                   ; .. (0FDH = IBM PCjr)
        CMP     BYTE PTR ES:[0EH],0FFH  ; .. (0FFH = IBM PC compatible)
	JE	Setup1			; ..
        CMP     BYTE PTR ES:[0EH],0FDH  ; ..
        JE      Setup2                  ; ..

	MOV	DX,OFFSET NoTandy	;Computer not an IBM PC compatible
	JMP	NoInstall		; ..
Setup1:
	MOV	AX,0FC00H		;Check byte at FC00:0000 for 21H
	MOV	ES,AX			; .. (Tandy 1000 unique)
	CMP	BYTE PTR ES:[0],21H	; ..
	JE	Setup2			; ..
	MOV	DX,OFFSET NoTandy	;Computer is not a Tandy 1000 series
	JMP	NoInstall		; ..

;See if program is already resident
SetUp2:
	MOV	AX,0EE80H	;See if this program is already resident
	MOV	BX,0		; ..
	INT	10H		; ..
	CMP	AX,0DEADH	;Test for return value
	JNE	SetUp3		; ..
	MOV	DX,OFFSET InstallError ;Print error message and exit
	JMP	NoInstall	       ; ..
SetUp3:
	MOV	AH,9		     ;Print opening message
	MOV	DX,OFFSET InstallMsg ; ..
	INT	21H		     ; ..

;Install new interrupt vectors and remain resident
	CLI			;Set the 8253 timer chip to normal speed
	CALL	SetTimer	; ..
	STI			; ..
	MOV	AX,3510H	;Get vector address for INT 10H
	INT	21H		; ..
	MOV	[INT10+2],ES	;Save vector address
	MOV	[INT10],BX	; ..
	MOV	AX,3508H	;Get vector address for INT 08
	INT	21H		; ..
	MOV	[INT08+2],ES	;Save vector address
	MOV	[INT08],BX	; ..
	MOV	AX,2510H	   ;Change video BIOS vector to new address
	MOV	DX,OFFSET TandySnd ; ..
	INT	21H		   ; ..
	MOV	AX,2508H	;Change BIOS timer vector to new address
	MOV	DX,OFFSET Timer ; ..
	INT	21H		; ..
	MOV	DX,OFFSET Setup ;Setup paragraph count for terminate
	MOV	CL,4		; ..
	SHR	DX,CL		; ..
	INC	DX		; ..
	MOV	AX,3100H	;Terminate and Stay Resident
	INT	21H		; ..

;Print error message and exit
NoInstall:
	MOV	AH,9		;Print error message and exit
	INT	21H		; ..
	MOV	AX,4CFFH	; ..
	INT	21H		; ..

InstallMsg:
	DB	13,10
        DB      13,10,'TANDYSND 1.2 Copyright (C) 1991-2023 by Joseph A. Albrecht'
        DB      13,10
        DB      13,10,'             Tandy 1000/PCjr Sound Extensions'
        DB      13,10,'           Texas Instruments SN76496 Sound Chip'
	DB	13,10
	DB	'$'

InstallError:
	DB	7
	DB	13,10
	DB	'TANDSND is already installed!'
	DB	13,10
	DB	'$'

NoTandy:
	DB	7
	DB	13,10
        DB      'TANDYSND only runs on a Tandy 1000/PCjr!'
	DB	13,10
	DB	'$'

Code	ENDS

	END	Start

